"""
Runner for Takt
"""
import argparse
import os

from nasse.logging import log, logger

logger.config.name = "takt"
DEFAULT_ENV = "TAKT_DISCORD_BOT_TOKEN"


def entry():
    """the CLI entrypoint"""
    parser = argparse.ArgumentParser("takt", description="Power your discord calls with music 🎐")
    parser.add_argument("--version", "-v", action="version", version="1.0")
    parser.add_argument("--token", "-t", action="store",
                        help=f"the discord bot token to use (the `{DEFAULT_ENV}` env var is used by default)", required=False, default=None)
    parser.add_argument("--prefix", "--command-prefix", "-p", action="store",
                        help="the prefix for the different commands", default="+", required=False)
    parser.add_argument("--debug", "-d", action="store_true", help="enabling debug logging")
    parser.add_argument("--cooldown", "-c", action="store", type=float, help="the cooldown between each command (per user) (in sec.)", default=1)

    args = parser.parse_args()
    from takt import config

    logger.config.debug = args.debug

    config.COMMAND_PREFIX = args.prefix
    config.DEBUG_MODE = args.debug
    config.COOLDOWN = args.cooldown

    from takt import takt
    from takt.receivers import client as _  # loading the receivers
    from takt.receivers import slash as _

    log("Starting up...")
    takt.client.run(args.token or os.environ[DEFAULT_ENV])


if __name__ == '__main__':
    entry()



"""
error.py

The error handler for Takt.
"""
from nasse.logging import LoggingLevel, log, logger

from discord.ext import commands

from takt.config import DEBUG_MODE
from takt.exceptions import TaktException


async def error_handler(context, error: commands.CommandInvokeError):
    if DEBUG_MODE:
        logger.print_exception()
    if isinstance(error, commands.CommandInvokeError):
        error = error.original
    if isinstance(error, TaktException):
        await context.send(error.MESSAGE.format(mention=context.author.mention))
    elif isinstance(error, commands.CommandNotFound):
        pass
    else:
        log("An unknown error occured: {name} {error}".format(
            name=error.__class__.__name__, error=str(error)), level=LoggingLevel.ERROR)
        await context.send(TaktException.MESSAGE.format(mention=context.author.mention))



"""receives plain text commands from discord"""
from discord.ext import commands

from takt.bot import client
from takt.takt import (clear, connected, help, latency, loop, looping, pause,
                       paused, play, playing, queue, resume, skip, stop)


@client.command(name="play", pass_context=True, aliases=["p", "start"])
async def play_receiver(context: commands.Context, *args):
    await play(context, " ".join(args))


@client.command(name="pause", pass_context=True)
async def pause_receiver(context: commands.Context):
    await pause(context)


@client.command(name="resume", pass_context=True, aliases=["continue", "unpause"])
async def resume_receiver(context: commands.Context):
    await resume(context)


@client.command(name="stop", pass_context=True)
async def stop_receiver(context: commands.Context):
    await stop(context)


@client.command(name="playing", pass_context=True, aliases=["now", "now-playing", "nowplaying", "np"])
async def playing_receiver(context: commands.Context):
    await playing(context)


@client.command(name="paused", pass_context=True, aliases=["ispaused"])
async def paused_receiver(context: commands.Context):
    await paused(context)


@client.command(name="connected", pass_context=True, aliases=["isconnected"])
async def connected_receiver(context: commands.Context):
    await connected(context)


@client.command(name="latency", pass_context=True)
async def latency_receiver(context: commands.Context):
    await latency(context)


@client.command(name="queue", pass_context=True, aliases=["q", "list", "list-queue", "listqueue", "song-queue", "songs", "musics", "songqueue"])
async def queue_receiver(context: commands.Context):
    await queue(context)


@client.command(name="skip", pass_context=True, aliases=["s", "next"])
async def skip_receiver(context: commands.Context):
    await skip(context)


@client.command(name="clear", pass_context=True)
async def clear_receiver(context: commands.Context):
    await clear(context)


@client.command(name="loop", pass_context=True, aliases=["repeat", "repeat-song", "repeatsong"])
async def loop_receiver(context: commands.Context, *args):
    await loop(context, " ".join(args))


@client.command(name="looping", pass_context=True, aliases=["is-looping", "islooping"])
async def looping_receiver(context: commands.Context):
    await looping(context)

client.remove_command("help")


@client.command(name="help", pass_context=True, aliases=["h", "commands", "command", "cmds", "cmd"])
async def help_receiver(context: commands.Context, *args):
    await help(context, " ".join(args))



"""
bot.py

Initialization of Takt and event registrations
"""
import discord
from discord.ext import commands
from discord_slash import SlashCommand
from nasse.logging import LoggingLevel, log

from takt.config import COMMAND_PREFIX
from takt.error import error_handler


async def prefix(bot, msg):
    """Defines the prefix for the bot"""
    return commands.when_mentioned_or(COMMAND_PREFIX)(bot, msg)

# DEFINING CLIENT/BOT AND ITS PREFIX
log("Defining client and slash client")
log(f"The defined prefix is '{COMMAND_PREFIX}' (or mention)")
client = commands.Bot(command_prefix=prefix, case_insensitive=True, intents=discord.Intents.default())
slash = SlashCommand(client, sync_commands=True)


@client.event
async def on_ready():
    # GAME ACTIVITY
    await client.change_presence(activity=discord.Game(name=f'{COMMAND_PREFIX}help'))
    log("Takt is ready", level=LoggingLevel.INFO)


@client.event
async def on_connect():
    log(f"Takt is connected to Discord (latency: {round(client.latency * 1000, 2)} ms)", level=LoggingLevel.INFO)


@client.event
async def on_disconnect():
    log(f"Takt got disconnected from Discord", level=LoggingLevel.INFO)


@client.event
async def on_command_error(context, exception):
    await error_handler(context, exception)



"""receives slash commands from discord"""
from discord_slash.model import SlashCommandOptionType
from discord_slash.utils.manage_commands import create_option

from takt.bot import slash
from takt.takt import (clear, connected, latency, loop, looping, pause, paused,
                       play, playing, queue, resume, skip, stop)


@slash.slash(name="play",
             description="Plays the given URL in the connected voice channel",
             options=[
                 create_option(
                     name="link",
                     description="The URL or a search term of the song to play",
                     option_type=SlashCommandOptionType.STRING,
                     required=True
                 )
             ])
async def play_receiver_slash(context, link: str):
    await play(context=context, link=link)


@slash.slash(name="pause",
             description="Pauses the currently playing song")
async def pause_receiver_slash(context):
    await pause(context=context)


@slash.slash(name="resume", description="Resumes the current song")
async def resume_receiver_slash(context):
    await resume(context=context)


@slash.slash(name="stop", description="Stops the current song")
async def stop_receiver_slash(context):
    await stop(context=context)


@slash.slash(name="playing", description="Shows the currently playing song")
async def playing_receiver_slash(context):
    await playing(context=context)


@slash.slash(name="queue", description="Shows the current queue")
async def queue_receiver_slash(context):
    await queue(context=context)


@slash.slash(name="skip", description="Skips the current song")
async def skip_receiver_slash(context):
    await skip(context=context)


@slash.slash(name="clear", description="Clears the current queue")
async def clear_receiver_slash(context):
    await clear(context=context)


@slash.slash(name="loop", description="Loops the current song")
async def loop_receiver_slash(context):
    await loop(context=context)



"""
exceptions.py

Exceptions models for Takt.
"""

from nasse.exceptions import NasseException


class TaktException(NasseException):
    # NasseException already handles the logging
    MESSAGE = "{mention} ❌ An error occured while processing your request"

    def __init__(self, message: str = None, *args: object) -> None:
        super().__init__(message=message, *args)


class NotInVoiceChannel(TaktException):
    MESSAGE = "{mention} ❌ You do not seem to be in a voice channel"


class NoVoiceClient(TaktException):
    MESSAGE = "{mention} ❌ It seems that Takt is not connected to any voice channel"


class DownloadError(TaktException):
    MESSAGE = "{mention} ❌ An error occured while downloading the audio. This might be because the given link is not supported."



"""the core"""
import asyncio
from time import time
from datetime import timedelta

import discord
from discord.ext import commands
from nasse.logging import log, LoggingLevel

from takt.config import COMMAND_PREFIX, COOLDOWN
from takt.audio import TaktAudioPlayer
from takt.bot import \
    client  # to redirect the import outside (and at the same time load takt)
from takt.exceptions import NotInVoiceChannel, NoVoiceClient

SERVERS = {}


def decorate(func):
    async def wrapper(context: commands.Context, *args, **kwargs):
        if context.guild.id not in SERVERS:
            SERVERS[context.guild.id] = {}
        if "RATE" not in SERVERS[context.guild.id]:
            SERVERS[context.guild.id]["RATE"] = {}
        distance = time() - SERVERS[context.guild.id]["RATE"].get(context.author.id, 0)
        if distance < COOLDOWN:
            log(f"`{context.author.name}` is rate limited for now ({COOLDOWN - distance} seconds remaining)", level=LoggingLevel.INFO)
            return await context.send(f"🏮 {context.author.mention} You are rate limited for now ({round(COOLDOWN - distance, 2)} seconds remaining)")
        result = await func(context, *args, **kwargs)
        SERVERS[context.guild.id]["RATE"][context.author.id] = time()
        return result
    return wrapper


def human_format(number: int):
    magnitude = 0
    while abs(number) >= 1000:
        magnitude += 1
        number /= 1000
    return f"{round(number, 2)}{['', 'K', 'M', 'B', 'T', 'P'][magnitude]}"


def create_end_event(context: commands.Context, loop: asyncio.events.AbstractEventLoop):
    def on_ended():
        if context.voice_client:
            context.voice_client.stop()
            log("Song ended")
            if context.guild.id not in SERVERS:
                log("No queue for current guild")
                return
            QUEUE = SERVERS[context.guild.id].get("QUEUE", [])
            try:
                now_playing = QUEUE.pop(0)
            except IndexError:
                log("Queue is empty")
                asyncio.run_coroutine_threadsafe(context.voice_client.disconnect(), loop)
                return

            if SERVERS[context.guild.id].get("LOOP", False):  # loop mechanism
                new = now_playing.new()
                new.on_ended = create_end_event(context, loop)
                QUEUE.append(new)

            if len(QUEUE) >= 1:
                log("More than one song in queue")
                log(f"Playing {QUEUE[0]}")
                context.voice_client.play(QUEUE[0])
            else:
                log("No more songs in queue")
                asyncio.run_coroutine_threadsafe(context.voice_client.disconnect(), loop)
                asyncio.run_coroutine_threadsafe(context.send("🛫 No more songs in queue, exiting..."), loop)
    return on_ended


@decorate
async def play(context: commands.Context, link: str):
    if context.author.voice is None:
        raise NotInVoiceChannel

    if context.voice_client is None:  # not connected yet to any voice channel
        log("Connecting to voice channel")
        await context.author.voice.channel.connect()
    else:
        log("Moving to new channel")
        await context.voice_client.move_to(context.author.voice.channel)
    # context.voice_client.stop()

    player = await TaktAudioPlayer.open(link)
    loop = asyncio.get_event_loop()
    player.on_ended = create_end_event(context, loop)

    if context.guild.id not in SERVERS:
        SERVERS[context.guild.id] = {}

    try:
        SERVERS[context.guild.id]["QUEUE"].append(player)
    except Exception:
        SERVERS[context.guild.id]["QUEUE"] = [player]

    if not context.voice_client.is_playing():
        context.voice_client.play(player)
        log(f"Playing {player}")
        await context.send(f'{context.author.mention} | ✨ Playing **"{player.title}"**')
    else:
        log(f"Adding {player} to queue")
        await context.send(f'{context.author.mention} | 🎏 Added **"{player.title}"** to the queue')


@decorate
async def pause(context: commands.Context):
    if context.voice_client is None:
        raise NoVoiceClient
    log("Pausing")
    context.voice_client.pause()
    await context.send(f"{context.author.mention} Paused")


@decorate
async def resume(context: commands.Context):
    if context.voice_client is None:
        raise NoVoiceClient
    if context.voice_client.is_playing():
        await context.send(f"{context.author.mention} Already playing")
        return
    log("Resuming")
    context.voice_client.resume()
    await context.send(f"{context.author.mention} Resumed")


@decorate
async def skip(context: commands.Context):
    if context.voice_client is None:
        raise NoVoiceClient
    log("Skipping")
    loop = asyncio.get_event_loop()
    create_end_event(context, loop)()
    await context.send(f"{context.author.mention} Skipped the current song!")


@decorate
async def loop(context: commands.Context, loop: str = None):
    if context.guild.id not in SERVERS:
        SERVERS[context.guild.id] = {}
    if loop is not None:
        loop = str(loop).lower().replace(' ', '')
        if loop in ("yes", "1", "true", "enable", "active", "activate"):
            loop = True
        elif loop in ("no", "0", "false", "disable", "inactive", "deactivate"):
            loop = False
        else:
            await context.send(f"{context.author.mention} Invalid argument `{loop}`")
            return
    else:
        loop = not SERVERS.get(context.guild.id, {}).get("LOOP", False)

    if loop:
        SERVERS[context.guild.id]["LOOP"] = True
        await context.send(f"{context.author.mention} | 🔁 Loop enabled!")
    else:
        SERVERS[context.guild.id]["LOOP"] = False
        await context.send(f"{context.author.mention} | 🏮 Loop disabled")


@decorate
async def looping(context: commands.Context):
    if SERVERS.get(context.guild.id, {}).get("LOOP", False):
        await context.send(f"{context.author.mention} | 🔁 Loop is ON!")
    else:
        await context.send(f"{context.author.mention} | 🏮 Loop is OFF!")


@decorate
async def queue(context: commands.Context):
    log(f"Sending the current guild queue (guild: {context.guild.name} | {context.guild.id})")
    if len(SERVERS.get(context.guild.id, {}).get("QUEUE", [])) == 0:
        await context.send(f"{context.author.mention} The queue is empty")
        return
    embed = discord.Embed(title='Current Queue',
                          colour=discord.Colour.blue())
    embed.add_field(name='Queue', value="\n".join(f"{i}. {player.title}" for i, player in enumerate(SERVERS[context.guild.id]["QUEUE"], 1)))
    length = len(SERVERS[context.guild.id]['QUEUE'])
    footer = f"{length} song{'s' if length > 1 else ''} in queue"
    if SERVERS.get(context.guild.id, {}).get("LOOP", False):
        footer += "\n🔁 Loop enabled"
    embed.set_footer(text=footer)

    await context.send(embed=embed)


@decorate
async def clear(context: commands.Context):
    log("Clearing the queue and looping status")
    SERVERS.pop(context.guild.id, None)
    await context.send(f"{context.author.mention} Cleared the queue")


@decorate
async def stop(context: commands.Context):
    if context.voice_client is None:
        raise NoVoiceClient
    log("Stopping and clearing")
    SERVERS.pop(context.guild.id, None)
    context.voice_client.stop()
    await context.voice_client.disconnect()
    await context.send(f"{context.author.mention} Stopped the music!")


@decorate
async def playing(context: commands.Context):
    if context.voice_client is None:
        raise NoVoiceClient
    if context.voice_client.is_playing():
        log("Creating and sending the 'Now Playing' embed")
        if context.voice_client.source.url:
            embed = discord.Embed(title='Now playing', colour=discord.Colour.blue(), url=context.voice_client.source.url)
        else:
            embed = discord.Embed(title='Now playing', colour=discord.Colour.blue())
        if context.voice_client.source.thumbnail:
            embed.set_thumbnail(url=context.voice_client.source.thumbnail)

        embed.add_field(name='Title', value=context.voice_client.source.title)
        if context.voice_client.source.duration:
            embed.add_field(
                name='Time', value=f"{timedelta(seconds=int(context.voice_client.source.played / 1000))}/{timedelta(seconds=int(context.voice_client.source.duration))}")
        else:
            embed.add_field(name='Time', value=str(timedelta(seconds=int(context.voice_client.source.played / 1000))))
        if context.voice_client.source.view_count:
            value = f"Views: {human_format(context.voice_client.source.view_count)}"
            if context.voice_client.source.like_count:
                value += f"\nLikes: {human_format(context.voice_client.source.like_count)}"
            embed.add_field(name='Counter', value=value)
        elif context.voice_client.source.like_count:
            embed.add_field(name='Counter', value=f"Likes: {human_format(context.voice_client.source.like_count)}")
        if context.voice_client.source.channel:
            embed.add_field(name='Channel', value=context.voice_client.source.channel)
        # embed.add_field(name='Description', value=context.voice_client.source.description)
        if context.voice_client.source.upload_date:
            embed.add_field(name='Uploaded', value=context.voice_client.source.upload_date.strftime("%d/%m/%Y"))

        if SERVERS.get(context.guild.id, {}).get("LOOP", False):
            embed.add_field(name="Options", value="🔁 Loop enabled")

        await context.send(f'{context.author.mention} Now playing: **{context.voice_client.source.title}**', embed=embed)
    else:
        await context.send(f"{context.author.mention} Nope, nothing is being played")


@decorate
async def paused(context: commands.Context):
    if context.voice_client is None:
        raise NoVoiceClient
    if context.voice_client.is_paused():
        await context.send(f"{context.author.mention} It seems that the music is paused")
    else:
        await context.send(f"{context.author.mention} Nope it seems that no music is paused right now")


@decorate
async def connected(context: commands.Context):
    if context.voice_client is None:
        raise NoVoiceClient
    if context.voice_client.is_connected():
        await context.send(f"{context.author.mention} Yes I am connected")
    else:
        await context.send(f"{context.author.mention} Nope I'm not connected")


@decorate
async def latency(context: commands.Context):
    if context.voice_client is None:
        raise NoVoiceClient
    log(f"Bot Latency: {client.latency}")
    log(f"Voice Latency: {round(context.voice_client.latency * 1000, 4)} (average: {round(context.voice_client.average_latency * 1000, 4)})")
    await context.send(f"{context.author.mention} The current latency is **{round(context.voice_client.latency * 1000, 2)}ms** (average: {round(context.voice_client.average_latency * 1000, 2)}ms)")


HELP = {
    "play <search term>": {
        "msg": "Searches on YouTube and plays the first result",
        "example": "play Bad Apple",
        "aliases": ["p <search term>", "start <search term>"]
    },
    "play <link>": {
        "msg": "Play the given link",
        "example": "play https://www.youtube.com/watch?v=dQw4w9WgXcQ",
        "aliases": ["p <link>", "start <link>"]
    },
    "playing": {
        "msg": "Show the current song playing",
        "aliases": ["now", "now-playing", "nowplaying", "np"]
    },
    "pause": {
        "msg": "Pause the current song"
    },
    "resume": {
        "msg": "Resume the current song",
        "aliases": ["continue", "unpause"]
    },
    "skip": {
        "msg": "Skip the current song",
        "aliases": ["s", "next"]
    },
    "queue": {
        "msg": "Show the current queue",
        "aliases": ["q", "list", "list-queue", "listqueue", "song-queue", "songs", "musics", "songqueue"]
    },
    "clear": {
        "msg": "Clear the queue"
    },
    "stop": {
        "msg": "Stop the music and clear the queue"
    },
    "loop": {
        "msg": "Toggle looping",
        "aliases": ["repeat", "repeat-song", "repeatsong"]
    },
    "loop <true/false>": {
        "msg": "Enable or disable looping",
        "example": "loop true"
    },
    "looping": {
        "msg": "Show the current loop status",
        "aliases": ["is-looping", "islooping"]
    },
    "latency": {
        "msg": "Show the current latency"
    },
    "connected": {
        "msg": "Show if the bot is connected to a voice channel",
        "aliases": ["isconnected"]
    },
    "help": {
        "msg": "Show this message",
        "aliases": ["h", "commands", "command", "cmds", "cmd"]
    },
    "help <command>": {
        "msg": "Show specific command's help",
        "example": "help play"
    }
}

ALIASES = {}
for cmd, data in HELP.items():
    for alias in data.get("aliases", []):
        a = str(alias).replace(" ", "").lower()
        ALIASES[a] = cmd
        ALIASES[a.split("<")[0]] = cmd
    c = str(cmd).replace(" ", "").lower()
    ALIASES[c] = cmd
    ALIASES[c.split("<")[0]] = cmd


@decorate
async def help(context: commands.Context, command: str = ""):
    embed = discord.Embed(title='🏮 Help Center', colour=discord.Colour.blue())
    embed.set_author(name=f"Requested by {context.author}")
    embed.set_footer(text="© Takt by Anise")
    command = str(command).replace(" ", "").lower()
    if command == "":
        embed.add_field(name='Commands', value="\n".join(
            f"`{COMMAND_PREFIX}{cmd}`{': ' + str(data['msg']) if 'msg' in data else ''}" for cmd, data in HELP.items()))
        await context.send(embed=embed)
        return
    if command.startswith(COMMAND_PREFIX):
        command = command[1:]
    if command not in ALIASES:
        await context.send(f"{context.author.mention} I don't know that command")
        return
    command = ALIASES[command]
    data = HELP[command]
    embed.title = f"🏮 Help Center — `{COMMAND_PREFIX}{command}`"
    if 'msg' in data:
        embed.add_field(name="Description", value=data['msg'], inline=False)
    if 'example' in data:
        embed.add_field(name="Example", value="**" + COMMAND_PREFIX + data['example'] + "**", inline=False)
    if 'aliases' in data:
        embed.add_field(name="Aliases", value=", ".join(f"`{COMMAND_PREFIX}{alias}`" for alias in data['aliases']), inline=False)
    await context.send(embed=embed)



"""manages audio on takt"""
from datetime import datetime

from discord.player import FFmpegOpusAudio
from nasse.logging import log
from nasse.utils.types import URL
from youtube_dl import YoutubeDL as YouTubeDL
from yt_dlp import YoutubeDL as YouTubeDL_P

from takt.exceptions import DownloadError


class TaktAudioPlayer(FFmpegOpusAudio):
    """
    Takt's custom discord audio player
    """

    def __init__(self, info, *, bitrate=128, volume: float = 0.1, codec="nocopy", executable='ffmpeg', pipe=False, stderr=None, before_options="-reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5", options='-vn'):
        self.info = dict(info)
        self.source = self.info["url"]
        self.played = 0

        # infos
        self.url = self.info.get("webpage_url", None)
        self.title = self.info.get("title", "(Unknown Title)")
        self.description = self.info.get("description", None)
        self.duration = self.info.get("duration", None)  # could have my own probe
        self.view_count = self.info.get("view_count", None)
        self.like_count = self.info.get("like_count", None)
        if "upload_date" in self.info:
            self.upload_date = datetime.strptime(self.info["upload_date"], "%Y%m%d")
        else:
            self.upload_date = None
        self.channel = self.info.get("uploader", None)
        self.channel_link = self.info.get("uploader", None)
        self.thumbnail = self.info.get("thumbnail", None)

        # other
        self.bitrate = bitrate
        self.volume = volume
        self.codec = codec
        self.executable = executable
        self.pipe = pipe
        self.stderr = stderr
        self.before_options = before_options
        self.options = options

        log(f"Setting the volume to {volume}")
        options += f' -filter:a "volume={volume}"'
        super().__init__(self.source, bitrate=bitrate, codec=codec, executable=executable,
                         pipe=pipe, stderr=stderr, before_options=before_options, options=options)

        log(f"Created {self}")

    async def open(link, *, bitrate=128, volume: float = 0.2, codec="nocopy", executable='ffmpeg', pipe=False, stderr=None, before_options="-reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5", options='-vn'):
        log(f"Creating a new TaktAudioPlayer with {link}")
        info = None
        for WORKER in (YouTubeDL_P, YouTubeDL):
            try:
                with WORKER({
                    "format": "bestaudio",
                    "noplaylist": "True"
                }) as worker:
                    if URL.REGEX and URL.REGEX.match(link):
                        info = worker.extract_info(link, download=False)
                    else:
                        info = worker.extract_info(f"ytsearch:{link}", download=False)["entries"][0]
                log("Found info")
                break
            except Exception:
                continue
        if info is None:
            raise DownloadError

        # i need to change the codec to reencode the audio
        _, bitrate = await FFmpegOpusAudio.probe(info["url"])
        log(f"Found Bitrate: {bitrate}")
        return TaktAudioPlayer(info, bitrate=bitrate, volume=volume, codec=codec, executable=executable, pipe=pipe, stderr=stderr, before_options=before_options, options=options)

    def on_ended(self) -> bool:
        log(f"{self} has ended")

    def read(self):
        data = super().read()
        if data != b"":
            self.played += 20
        else:
            self.on_ended()
        return data

    def __repr__(self) -> str:
        return "TaktAudio(title={}, duration={}, played={} sec)".format(self.title, self.duration, round(self.played / 1000, 2))

    def new(self):
        return TaktAudioPlayer(self.info, bitrate=self.bitrate, volume=self.volume, codec=self.codec, executable=self.executable, pipe=self.pipe, stderr=self.stderr, before_options=self.before_options, options=self.options)
